/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */
/***************************************************************************
 *            e-cal-backend-kolab.c
 *
 *  Thu Jun 10 18:25:26 2010
 *  Copyright  2010  Christian Hilberg
 *  <hilberg@kernelconcepts.de>
 *  and Silvan Marco Fin <silvan@kernelconcepts.de> in 2011
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

/*----------------------------------------------------------------------------*/

#include <glib.h>
#include <gio/gio.h>

#include <libedata-cal/e-cal-backend-cache.h>
#include <libecal/e-cal.h>

#include <libical/ical.h>

#include <libekolab/kolab-mail-access.h>
#include <libekolab/kolab-settings-handler.h>
#include <libekolab/kolab-mail-handle.h>
#include <libekolabutil/kolab-util-http.h>
#include <libekolabutil/kolab-util-glib.h>
#include <libekolabutil/kolab-util-camel.h>
#include <libekolabutil/kolab-util-cal-freebusy.h>

#include "kolab-util-calendar.h"
#include "kolab-util-calendar-cache.h"
#include "e-cal-backend-kolab.h"

/*----------------------------------------------------------------------------*/
/* table of KolabMailAccess objects */

/* static GHashTable *koma_objects = NULL; */

/*----------------------------------------------------------------------------*/

#define E_CAL_BACKEND_KOLAB_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), E_TYPE_CAL_BACKEND_KOLAB, ECalBackendKolabPrivate))

G_DEFINE_TYPE (ECalBackendKolab, e_cal_backend_kolab, E_TYPE_CAL_BACKEND_SYNC)

/* private structures for Kolab calendar class ********************************/

/* TODO check what is appropriate for Kolab here */

typedef struct {
	GCond *cond;
	GMutex *mutex;
	gboolean exit;
} SyncDelta;

/* Private part of the ECalBackendKolab structure */
typedef struct _ECalBackendKolabPrivate ECalBackendKolabPrivate;
struct _ECalBackendKolabPrivate {
	GHashTable		*koma_table;
	CalMode			cal_mode;
	KolabMailAccess		*cal_koma;
	ECalBackendCache	*cal_cache;
	gchar			*user_email;
	ECalComponent		*default_zone;
	ECalSourceType		source_type;
	gboolean		already_opened;
	gchar *			cal_uri;
};

/*----------------------------------------------------------------------------*/
/* internal statics */

/**
 * e_cal_backend_kolab_open:
 * @backend: An ECalBackendSync object.
 * @cal: An EDataCal object.
 * @cancellable: A cancellation stack.
 * @only_if_exists: Whether to open the calendar if and only if it already exists
 * or just create it when it does not exist.
 * @err: A GError placeholder.
 *
 * Opens a calendar backend with data from a calendar stored at the specified URI.
 */
static void
e_cal_backend_kolab_open (ECalBackendSync *backend,
                          EDataCal *cal,
                          GCancellable *cancellable,
                          gboolean only_if_exists,
                          GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	(void)only_if_exists; /* FIXME */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

#if 0 /* FIXME old */
	/* to trigger eds to pass along username and password, set the property
	 * "auth" to "true" in the source setting:
	 * <property name="auth" value="true"/>
	 * <property name="username" value="..."/>
	 */

	if (priv->already_opened == TRUE) return status;

	esource = e_cal_backend_get_source (E_CAL_BACKEND (backend));
	/* Getting the esource from the cal-Object did not work as expected? */

	icalkind = e_cal_backend_get_kind (E_CAL_BACKEND (kolab));
	/* TODO: this has to be set according to the get_kind() method */
	switch (icalkind) {
	case ICAL_VEVENT_COMPONENT:
		priv->source_type = E_CAL_SOURCE_TYPE_EVENT;
		break;
	case ICAL_VTODO_COMPONENT:
		priv->source_type = E_CAL_SOURCE_TYPE_TODO;
		break;
	case ICAL_VJOURNAL_COMPONENT:
		priv->source_type = E_CAL_SOURCE_TYPE_JOURNAL;
		break;
	default:
		g_error ("%s()[%u]: Unknown Type used in e-cal-backend-kolab initialization!",
		         __func__, __LINE__);
	}

	kolab_util_glib_init ();
	kolab_util_http_init ();
	/* libcamel
	 * Curl init may configure the underlying SSL lib,
	 * but as far as SSL goes, we want Camel to rule here
	 * TODO check whether Camel session needs to be initialized before or after libcurl.
	 */
	ok = kolab_util_camel_init (&error);
	if (! ok) {
		status = kolab_util_calendar_map_error (error);
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__, error->message);
		g_error_free (error);
		error = NULL;
	}

	priv->cal_uri = e_source_get_uri(esource);

	cex = camel_exception_new ();
	c_url = camel_url_new (priv->cal_uri, cex);
	if (c_url == NULL) {
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__,
		           camel_exception_get_description (cex));
		camel_exception_free (cex);
		status = GNOME_Evolution_Calendar_OtherError;
	}
	servername = g_strdup (c_url->host);
	camel_url_free (c_url);

	/* Initialize backend cache */
	if (priv->cal_cache != NULL) {
		g_object_unref (priv->cal_cache);
	}
	priv->cal_cache = e_cal_backend_cache_new (priv->cal_uri,
	                                           priv->source_type);
	ok = e_file_cache_clean (E_FILE_CACHE (priv->cal_cache));

	/* Prepare data from sync strategy property */
	prop_str = e_source_get_property (esource, KOLAB_SYNC_STRATEGY_PROP);
	sync_value = kolab_util_misc_sync_value_from_property (prop_str);
	sourcename = kolab_util_backend_get_relative_path_from_uri (priv->cal_uri);

	/* This was the easy part. Now try to get the KoMA objects figured out */
	user_at_server = g_strdup_printf ("%s@%s", username, servername);

	ok = g_hash_table_lookup_extended (priv->koma_table,
	                                   user_at_server,
	                                   (gpointer *) &tmp_key,
	                                   (gpointer *) &tmp_koma);
	if (ok) {
		/* There is already a KoMA instance for $servername. Use it and return */
		g_object_ref (tmp_koma);
		priv->cal_koma = tmp_koma;
		g_free (user_at_server);
		ksettings = kolab_mail_access_get_settings_handler (priv->cal_koma);
		kolab_util_backend_prepare_settings (ksettings,
		                                     NULL,
		                                     NULL,
		                                     NULL,
		                                     NULL,
		                                     sourcename,
		                                     &sync_value);
		goto notifications;
	}
	/* Otherwise we need to setup a new KoMA instance and a settings handler */

	ksettings = KOLAB_SETTINGS_HANDLER (g_object_new (KOLAB_TYPE_SETTINGS_HANDLER, NULL));
	(void) kolab_settings_handler_configure (ksettings,
	                                         KOLAB_FOLDER_CONTEXT_CALENDAR,
	                                         &error);
	if (error != NULL) {
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__, error->message);
		g_error_free (error);
		error = NULL;
	}
	(void) kolab_settings_handler_bringup (ksettings,
	                                       &error);
	if (error != NULL) {
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__, error->message);
		g_error_free (error);
		error = NULL;
	}
	kolab_util_backend_prepare_settings (ksettings,
	                                     esource,
	                                     servername,
	                                     username,
	                                     password,
	                                     sourcename,
	                                     &sync_value);

	priv->cal_koma = KOLAB_MAIL_ACCESS (g_object_new (KOLAB_TYPE_MAIL_ACCESS, NULL));
	g_object_add_toggle_ref (G_OBJECT (priv->cal_koma),
	                         kolab_util_backend_koma_table_cleanup_cb,
	                         priv->koma_table);
	g_hash_table_insert (priv->koma_table,
	                     user_at_server,
	                     priv->cal_koma);

	kolab_mail_access_configure (priv->cal_koma,
	                             ksettings,
	                             &error);
	g_object_unref (ksettings);
	if (error != NULL) {
		status = kolab_util_calendar_map_error (error);
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__, error->message);
		g_error_free (error);
		return status;
	}
	kolab_mail_access_bringup (priv->cal_koma,
	                           &error);
	if (error != NULL) {
		status = kolab_util_calendar_map_error (error);
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__, error->message);
		g_error_free (error);
		return status;
	}

 notifications:
	priv->already_opened = TRUE;
	e_cal_backend_kolab_set_mode (E_CAL_BACKEND (backend),
	                              priv->cal_mode);

	g_free (sourcename);
	g_free (servername);

	return status;
#endif
}

static void
e_cal_backend_kolab_remove (ECalBackendSync *backend,
                            EDataCal *cal,
                            GCancellable *cancellable,
                            GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

#if 0 /* FIXME old */
	if (priv->already_opened != FALSE) {
		g_object_unref (priv->cal_koma);
		priv->cal_koma = NULL;
	}
	priv->already_opened = FALSE;

	/* Remove backend cache */
	if (priv->cal_cache != NULL) {
		(void) e_file_cache_remove (E_FILE_CACHE (priv->cal_cache));
		g_object_unref (priv->cal_cache);
		priv->cal_cache = NULL;
	}

	return GNOME_Evolution_Calendar_Success;
#endif
}

static void
e_cal_backend_kolab_refresh (ECalBackendSync *backend,
                             EDataCal *cal,
                             GCancellable *cancellable,
                             GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);
}

static gboolean
e_cal_backend_kolab_get_backend_property (ECalBackendSync *backend,
                                          EDataCal *cal,
                                          GCancellable *cancellable,
                                          const gchar *prop_name,
                                          gchar **prop_value,
                                          GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	g_assert (prop_name != NULL);
	(void)prop_value; /* FIXME */ /* must be NULL? */
	g_return_val_if_fail (err == NULL || *err == NULL, FALSE);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

	return FALSE;
}

static gboolean
e_cal_backend_kolab_set_backend_property (ECalBackendSync *backend,
                                          EDataCal *cal,
                                          GCancellable *cancellable,
                                          const gchar *prop_name,
                                          const gchar *prop_value,
                                          GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	g_assert (prop_name != NULL);
	(void)prop_value; /* FIXME */ /* may be NULL? */
	g_return_val_if_fail (err == NULL || *err == NULL, FALSE);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

	return FALSE;
}

/**
 * e_cal_backend_kolab_get_object:
 * @backend: An ECalBackendSync object.
 * @cal: An EDataCal object.
 * @cancellable: A cancellation stack.
 * @uid: UID of the object to get.
 * @rid: Recurrence ID of the specific instance to get, or NULL if getting the
 * master object.
 * @calobject: Placeholder for returned object.
 * @err: GError placeholder.
 *
 * Queries a calendar backend for a calendar object based on its unique
 * identifier and its recurrence ID (if a recurrent appointment).
 */
static void
e_cal_backend_kolab_get_object (ECalBackendSync *backend,
                                EDataCal *cal,
                                GCancellable *cancellable,
                                const gchar *uid,
                                const gchar *rid,
                                gchar **calobj,
                                GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	g_assert (uid != NULL);
	(void)rid; /* FIXME */ /* may be NULL? */
	(void)calobj; /* FIXME */ /* must be NULL? */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

#if 0 /* FIXME old */
	(void) cal;
	(void) rid;
	ecalcomp = kolab_util_calendar_cache_get_object (priv->cal_cache,
	                                                 priv->cal_koma,
	                                                 priv->cal_uri,
	                                                 uid,
	                                                 FALSE,
	                                                 &error);
	if (error != NULL) {
		status = kolab_util_calendar_map_error (error);
		g_warning ("%s()[%u]: %s", __func__, __LINE__, error->message);
		g_error_free (error);
		return status;
	}
	if (ecalcomp != NULL) {
		*object = (gchar *) e_cal_component_get_as_string (ecalcomp);

		g_object_unref (ecalcomp);
	}
	else {
		*object = NULL;
	}

	return GNOME_Evolution_Calendar_Success;
#endif
}

static void
e_cal_backend_kolab_get_object_list (ECalBackendSync *backend,
                                     EDataCal *cal,
                                     GCancellable *cancellable,
                                     const gchar *sexp,
                                     GSList **calobjs,
                                     GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	(void)sexp; /* FIXME */ /* sexp may be NULL */
	(void)calobjs; /* FIXME */ /* must be NULL? */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

#if 0 /* FIXME old */
	(void) cal;

	sourcename = kolab_util_backend_get_relative_path_from_uri (priv->cal_uri);

	kolab_util_calendar_cache_update_on_query (priv->cal_cache,
	                                           priv->cal_koma,
	                                           query,
	                                           priv->cal_uri);
	uid_list = kolab_mail_access_query_uids (priv->cal_koma,
	                                         sourcename,
	                                         query,
	                                         &error);
	if (error != NULL) {
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__, error->message);
		g_error_free (error);
		error = NULL;
	}

	for (it = g_list_first (uid_list); it != NULL; it = g_list_next (it)) {
		gchar *uid = it->data;
		ECalComponent *ecalcomp = NULL;
		ecalcomp = kolab_util_calendar_cache_get_object (priv->cal_cache,
		                                                 priv->cal_koma,
		                                                 priv->cal_uri,
		                                                 uid,
		                                                 TRUE,
		                                                 &error);
		if (error != NULL) {
			g_warning ("%s()[%u]: %s",
			           __func__, __LINE__, error->message);
			g_error_free (error);
			error = NULL;
		}
		if (ecalcomp != NULL) {
			*objects = g_list_append (*objects,
			                          e_cal_component_get_as_string (ecalcomp));
			g_object_unref (ecalcomp);
		}
	}
	g_free (sourcename);

	kolab_util_glib_glist_free (uid_list);

	return GNOME_Evolution_Calendar_Success;
#endif
}

static void
e_cal_backend_kolab_get_free_busy (ECalBackendSync *backend,
                                   EDataCal *cal,
                                   GCancellable *cancellable,
                                   const GSList *users,
                                   time_t start,
                                   time_t end,
                                   GSList **freebusyobjs,
                                   GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	(void)users; /* FIXME */
	(void)start; /* FIXME */
	(void)end; /* FIXME */
	(void)freebusyobjs; /* FIXME */ /* must be NULL? */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

#if 0 /* FIXME old */
	(void) cal;
	(void) start;
	(void) end;

	ksettings = kolab_mail_access_get_settings_handler (priv->cal_koma);
	if (ksettings == NULL)
		return GNOME_Evolution_Calendar_OtherError;

	/* receive the xfb information */
	for (it = g_list_first(users); it != NULL; it = g_list_next (it)) {
		KolabHttpJob *job = NULL;
		ECalComponent *ecalcomp = NULL;

		job = kolab_util_calendar_retrieve_xfb (ksettings, (gchar *) it->data);
		if (job != NULL) {
			ecalcomp = kolabconv_cal_util_freebusy_ecalcomp_new_from_ics ((gchar *) (job->buffer->data),
			                                                              job->nbytes,
			                                                              &error);
			if (error != NULL) {
				g_warning ("%s", error->message);
				g_error_free (error);
				error = NULL;
				internal_state = FALSE;
			}
			kolab_util_http_job_free (job);
		}
		if (ecalcomp == NULL) {
			/* this is not an error, but there's no F/B
			 * information available.
			 */
			g_object_unref (ksettings);
			return GNOME_Evolution_Calendar_ObjectNotFound;
		}
		else {
			*freebusy = g_list_append (*freebusy,
			                           g_strdup (e_cal_component_get_as_string (ecalcomp)));
			g_object_unref (ecalcomp);
		}
	}
	g_object_unref (ksettings);
	if (internal_state == TRUE)
		return GNOME_Evolution_Calendar_Success;
	else
		return GNOME_Evolution_Calendar_OtherError;
#endif
}

/**
 * e_cal_backend_kolab_create_object:
 * @backend: (sync) kolab calendar backend object.
 * @cal: An EDataCal object.
 * @cancellable: A cancellation stack
 * @calobj: Contains the event to be created in the calendar as iCalendar string.
 * @uid: Return value of the created events uid.
 * @new_component: The newly created component, serialized as ical component string.
 * @err: A GError placeholder.
 *
 * Creates a new calendar entry in the Kolab calendar. This method gets called
 * when Evolution requests a new entry to be saved in the calendar.
 */
static void
e_cal_backend_kolab_create_object (ECalBackendSync *backend,
                                   EDataCal *cal,
                                   GCancellable *cancellable,
                                   const gchar *calobj,
                                   gchar **uid,
                                   ECalComponent **new_component,
                                   GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	(void)calobj; /* FIXME */
	(void)uid; /* FIXME */
	g_return_if_fail (new_component == NULL || *new_component == NULL);
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

#if 0 /* FIXME old */
	(void) cal;

	/* What do we have here? There is an ECalComponent, unable to hold
	 * timezone information and an calendar entry at the same time.
	 * So Kolab glue code expects two ECalComponents, one for the calendar entry
	 * one for the timezone.
	 */
	ecalcomp = e_cal_component_new_from_string (*calobj);
	if (ecalcomp == NULL) {
		g_warning (" + **calobj could not be parsed into ECalComponent: %s",
		           *calobj);
		return GNOME_Evolution_Calendar_InvalidObject;
	}
	tzcomp = kolab_util_calendar_cache_get_tz (priv->cal_cache,
	                                           ecalcomp);

	if (tzcomp == NULL) {
		gchar *tzid = NULL;
		tzid = kolab_util_calendar_get_tzid (ecalcomp,
		                                     E_CAL_COMPONENT_FIELD_DTSTART);
		/* If tzid is UTC, everything is fine */
		if (g_strcmp0 ("UTC", tzid) == 0) {
			g_free (tzid);
			tzid = NULL;
		}
		if (tzid != NULL) {
			/* TZID is set, but no timezone is available. Currently
			 * this may happen, if an object is copied to an yet unused/empty
			 * calendar. We do not want such crap in the kolab database.
			 */
			g_free (tzid);
			g_object_unref (ecalcomp);
			/* Don't ever expect to get to see the following message */
			e_data_cal_notify_error (cal, "TZID is set, but no timezone is available. "
			                         "Currently this may happen if an object is copied "
			                         "to an empty calendar.");
			return GNOME_Evolution_Calendar_InvalidObject;
		}
	}

	if (! kolab_util_calendar_cache_assure_uid_on_ecalcomponent (priv->cal_cache,
	                                                             priv->cal_koma,
	                                                             priv->cal_uri,
	                                                             ecalcomp,
	                                                             FALSE,
	                                                             &error)) {
		status = kolab_util_calendar_map_error (error);
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__, error->message);
		g_error_free (error);
		g_object_unref (ecalcomp);
		if (tzcomp != NULL)
			g_object_unref (tzcomp);
		return status;
	}

	kolab_util_calendar_store (ecalcomp,
	                           tzcomp,
	                           priv->default_zone,
	                           priv->cal_koma,
	                           priv->cal_uri,
	                           &error);
	if (error != NULL) {
		status = kolab_util_calendar_map_error (error);
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__, error->message);
		g_error_free (error);
		return status;
	}
	e_cal_component_get_uid (ecalcomp, &tmp_uid);
	*uid = g_strdup (tmp_uid);
	*calobj = e_cal_component_get_as_string (ecalcomp);

	if (tzcomp != NULL)
		g_object_unref (tzcomp);
	g_object_unref (ecalcomp);

	return GNOME_Evolution_Calendar_Success;
#endif
}

/**
 * e_cal_backend_kolab_modify_object:
 * @backend: An ECalBackendSync object.
 * @cal: An EDataCal object.
 * @cancellable: A cancellation stack.
 * @calobj: Object to be modified.
 * @mod: Type of modification to be done.
 * @old_component: Placeholder for returning the old object as it was stored on the
 * backend.
 * @new_component: Placeholder for returning the new object as it has been stored
 * on the backend.
 * @err: A GError placeholder.
 *
 * Requests the calendar backend to modify an existing object. If the object
 * does not exist on the calendar, it will be issued as new event.
 */
static void
e_cal_backend_kolab_modify_object (ECalBackendSync *backend,
                                   EDataCal *cal,
                                   GCancellable *cancellable,
                                   const gchar *calobj,
                                   CalObjModType mod,
                                   ECalComponent **old_component,
                                   ECalComponent **new_component,
                                   GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	(void)calobj; /* FIXME */
	(void)mod; /* FIXME */
	g_return_if_fail (old_component == NULL || *old_component == NULL); /* FIXME */
	g_return_if_fail (new_component == NULL || *new_component == NULL); /* FIXME */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

# if 0 /* FIXME old */
	(void) mod;
	(void) cal;

	/* calobj contains the modifications to be performed.
	 * ecalcomp is the ECalComponent created from calobj.
	 * scenarios:
	 *  a) ecalcomp is no part of a recurring event:
	 *     read oldobj by means of ecalcomps uid
	 *     store ecalcomp and we are done.
	 *  b) ecalcomp is a modification to all instances of a recurring event:
	 *     should be handled like (a)
	 *  c) ecalcomp is a single instance of a recurring event:
	 *     Kolab can not handle detached recurrences. Remove the instance
	 *     from the recurrence and create a new single event für ecalcomp.
	 */

	ecalcomp = e_cal_component_new_from_string (calobj);
	if (ecalcomp == NULL)
		return GNOME_Evolution_Calendar_InvalidObject;
	e_cal_component_get_uid (ecalcomp, &uid);

	oldcomp = kolab_util_calendar_cache_get_object (priv->cal_cache,
	                                                priv->cal_koma,
	                                                priv->cal_uri,
	                                                uid,
	                                                FALSE,
	                                                &error);
	if (error != NULL) {
		/* TODO: how to handle request to modify non existing objects? */
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__, error->message);
		g_error_free (error);
		g_object_unref (ecalcomp);
		g_object_unref (oldcomp);
		return GNOME_Evolution_Calendar_ObjectNotFound;
	}

	*old_object = e_cal_component_get_as_string (oldcomp);
	/* Is this correct? Perhaps _get_tz (..., oldcomp) is needed.
	 * This depends on wether evolution sends the timezone, if modification
	 * to the timezone has been made. In any other case the timezone
	 * information from ecalcomp and oldcomp should be all the same.
	 */
	ecaltz = kolab_util_calendar_cache_get_tz (priv->cal_cache,
	                                           ecalcomp);
	is_instance = e_cal_component_is_instance (ecalcomp);
	if (is_instance) {
		/* Handling of (c) */
		ECalComponent *newcomp = NULL;
		ECalComponent *oldtz = NULL;
		gchar *rid = NULL;
		gchar *notify_object = NULL;

		oldtz = kolab_util_calendar_cache_get_tz (priv->cal_cache,
		                                          oldcomp);
		rid = e_cal_component_get_recurid_as_string (ecalcomp);
		/* This removes the recurrence from ecalcomp. */
		e_cal_component_set_recurid (ecalcomp, NULL);
		/* This removes ecalcomp from the recurrence. */
		newcomp = kolab_util_calendar_cache_remove_instance (priv->cal_cache,
		                                                     CALOBJ_MOD_THIS,
		                                                     oldcomp,
		                                                     uid,
		                                                     rid);
		if (! kolab_util_calendar_cache_assure_uid_on_ecalcomponent (priv->cal_cache,
		                                                             priv->cal_koma,
		                                                             priv->cal_uri,
		                                                             ecalcomp,
		                                                             TRUE,
		                                                             &error)) {
			status = kolab_util_calendar_map_error (error);
			g_warning ("%s()[%u]: %s",
			           __func__, __LINE__, error->message);
			g_error_free (error);
			if (oldtz != NULL)
				g_object_unref (oldtz);
			g_object_unref (ecalcomp);
			if (ecaltz != NULL)
				g_object_unref (ecaltz);
			g_object_unref (newcomp);
			g_object_unref (oldcomp);
			g_free (*old_object);
			old_object = NULL;
			return status;
		}

		kolab_util_calendar_store (newcomp,
		                           oldtz,
		                           priv->default_zone,
		                           priv->cal_koma,
		                           priv->cal_uri,
		                           &error);
		if (error != NULL) {
			status = kolab_util_calendar_map_error (error);
			g_warning ("%s()[%u]: error : %s",
			           __func__, __LINE__, error->message);
			g_error_free (error);
			if (oldtz != NULL)
				g_object_unref (oldtz);
			g_object_unref (ecalcomp);
			if (ecaltz != NULL)
				g_object_unref (ecaltz);
			g_object_unref (newcomp);
			g_object_unref (oldcomp);
			g_free (*old_object);
			old_object = NULL;
			return status;
		}
		notify_object = e_cal_component_get_as_string (newcomp);
		e_cal_backend_notify_object_modified (E_CAL_BACKEND (kolab),
		                                      *old_object,
		                                      notify_object);
		g_free (notify_object);
		g_object_unref (newcomp);
		if (oldtz != NULL)
			g_object_unref (oldtz);
	}
	g_object_unref (oldcomp);
	kolab_util_calendar_store (ecalcomp,
	                           ecaltz,
	                           priv->default_zone,
	                           priv->cal_koma,
	                           priv->cal_uri,
	                           &error);
	if (error != NULL) {
		status = kolab_util_calendar_map_error (error);
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__, error->message);
		g_error_free (error);
		g_object_unref (ecalcomp);
		return status;
	}
	*new_object = g_strdup (e_cal_component_get_as_string (ecalcomp));
	if (is_instance) {
		e_cal_backend_notify_object_created (E_CAL_BACKEND (backend),
		                                     *new_object);
	}

	g_object_unref (ecalcomp);
	if (ecaltz != NULL)
		g_object_unref (ecaltz);

	return GNOME_Evolution_Calendar_Success;
#endif
}

/**
 * e_cal_backend_kolab_remove_object:
 * @backend: An ECalBackendSync object.
 * @cal: An EDataCal object.
 * @cancellable: A cancellation stack.
 * @uid: UID of the object to remove.
 * @rid: Recurrence ID of the instance to remove, or NULL if removing the
 * whole object.
 * @mod: Type of removal.
 * @old_component: Placeholder for returning the old object as it was stored on the
 * backend.
 * @new_component: Placeholder for returning the object after it has been modified (when
 * removing individual instances). If removing the whole object, this will be
 * NULL.
 * @err: A GError placeholder.
 *
 * Removes an (a part of a [recurring] ) object from a calendar backend. The
 * backend will notify all of its clients about the change.
 */
static void
e_cal_backend_kolab_remove_object (ECalBackendSync *backend,
                                   EDataCal *cal,
                                   GCancellable *cancellable,
                                   const gchar *uid,
                                   const gchar *rid,
                                   CalObjModType mod,
                                   ECalComponent **old_component,
                                   ECalComponent **new_component,
                                   GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	(void)uid; /* FIXME */
	(void)rid; /* FIXME */ /* may be NULL? */
	(void)mod; /* FIXME */
	g_return_if_fail (old_component == NULL || *old_component == NULL); /* FIXME */
	g_return_if_fail (new_component == NULL || *new_component == NULL); /* FIXME */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

#if 0 /* FIXME old */
	(void) cal;

	if (rid == NULL)
		mod = CALOBJ_MOD_ALL;

	oldcomp = kolab_util_calendar_cache_get_object (priv->cal_cache,
	                                                priv->cal_koma,
	                                                priv->cal_uri,
	                                                uid,
	                                                FALSE,
	                                                &error);
	if (error != NULL) {
		switch (error->code) {
		case KOLAB_BACKEND_ERROR_NOTFOUND:
			(void) e_cal_backend_cache_remove_component (priv->cal_cache,
			                                             uid,
			                                             NULL);
			/* Don't bother if this worked */
			id = g_new0 (ECalComponentId, 1);
			id->uid=g_strdup (uid);
			id->rid=g_strdup (rid);
			e_cal_backend_notify_object_removed (E_CAL_BACKEND (backend),
			                                     id,
			                                     NULL,
			                                     NULL);
			g_free (id);
			status = GNOME_Evolution_Calendar_Success;
			break;
		default:
			status = kolab_util_calendar_map_error (error);
			g_warning ("%s()[%u]: error (%s) code (%i)",
			           __func__, __LINE__, error->message, error->code);
		}
		g_error_free (error);
		return status;
	}

	*old_object = e_cal_component_get_as_string (oldcomp);
	id = e_cal_component_get_id (oldcomp);
	switch (mod) {
	case CALOBJ_MOD_THIS:
		oldtz = kolab_util_calendar_cache_get_tz (priv->cal_cache,
		                                          oldcomp);
		newcomp = kolab_util_calendar_cache_remove_instance (priv->cal_cache,
		                                                     CALOBJ_MOD_THIS,
		                                                     oldcomp,
		                                                     uid,
		                                                     rid);
		/* newcomp now contains an exception rule for the rid */
		kolab_util_calendar_store (newcomp,
		                           oldtz,
		                           priv->default_zone,
		                           priv->cal_koma,
		                           priv->cal_uri,
		                           &error);
		if (error != NULL) {
			status = kolab_util_calendar_map_error (error);
			g_warning ("%s()[%u]: error %s",
			           __func__, __LINE__, error->message);
			g_error_free (error);
			g_object_unref (newcomp);
			g_object_unref (oldcomp);
			if (oldtz != NULL)
				g_object_unref (oldtz);
			return status;
		}
		*object = e_cal_component_get_as_string (newcomp);
		g_object_unref (newcomp);
		break;
	case CALOBJ_MOD_THISANDPRIOR: /* not supported by ui */
		g_warning ("%s()[%u]: removing this and prior not supported.",
		           __func__, __LINE__);
		status = GNOME_Evolution_Calendar_UnsupportedMethod;
		break;
	case CALOBJ_MOD_THISANDFUTURE: /* not supported by ui */
		g_warning ("%s()[%u]: removing this and future not supported.",
		           __func__, __LINE__);
		status = GNOME_Evolution_Calendar_UnsupportedMethod;
		break;
	case CALOBJ_MOD_ALL:
	default:
		sourcename = kolab_util_backend_get_relative_path_from_uri (priv->cal_uri);
		ok = kolab_mail_access_delete_by_uid (priv->cal_koma,
		                                      uid,
		                                      sourcename,
		                                      &error);
		if (error != NULL) {
			status = kolab_util_calendar_map_error (error);
			g_warning ("%s()[%u]: error %s",
			           __func__, __LINE__, error->message);
			g_error_free (error);
			error = NULL;
		} else {
			/* Notification is not necessary. (?) */
			status = GNOME_Evolution_Calendar_Success;
		}
		(void) e_cal_backend_cache_remove_component (priv->cal_cache,
		                                             uid,
		                                             NULL);

		ksettings = kolab_mail_access_get_settings_handler (priv->cal_koma);
		g_object_unref (ksettings);
		*object = NULL;
	}
	g_object_unref (oldcomp);
	/* Do we need this? */

	ok = kolab_mail_access_source_fbtrigger_needed (priv->cal_koma,
	                                                sourcename,
	                                                &error);
	if (error != NULL)
		{
			status = kolab_util_calendar_map_error (error);
			g_warning ("%s()[%u]: %s",
			           __func__, __LINE__, error->message);
			g_error_free (error);
			error = NULL;
			goto cleanup;
		}
	if (ok) {
		ksettings = kolab_mail_access_get_settings_handler (priv->cal_koma);
		kolab_util_calendar_toggle_pfb_trigger (ksettings,
		                                        sourcename);
		g_object_unref (ksettings);
	}

	e_cal_backend_notify_object_removed (E_CAL_BACKEND (kolab),
	                                     id,
	                                     *old_object,
	                                     *object);
	e_cal_component_free_id (id);

 cleanup:
	g_free (sourcename);

	return status;
#endif
}

/**
 * e_cal_backend_kolab_receive_objects:
 * @backend: An ECalBackendSync object.
 * @cal: An EDataCal object.
 * @cancellable: A cancellation stack.
 * @calobj: iCalendar object to receive.
 * @err: A GError placeholder.
 *
 * Submits a (series of) calendar event(s) through iCalendar object(s).
 */
static void
e_cal_backend_kolab_receive_objects (ECalBackendSync *backend,
                                     EDataCal *cal,
                                     GCancellable *cancellable,
                                     const gchar *calobj,
                                     GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	(void)calobj; /* FIXME */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

#if 0 /* FIXME old */
	(void) cal;

	/* Workflow:
	 * Incoming calobj is supposed to be a vcalendar string.
	 * KolabMailHandle is the data structure we want to
	 * generate and KolabMailHandle takes an ECalComponent and a timezone
	 * (as ECalComponent). The task at hand: convert calobj into an
	 * icalcomponent ( #1 ). Afterwards process this icalcomponent and extract
	 * the relevant information an ECalComponent can handle ( #2 ). Then
	 * create the ECalComponent, generate a timezone and stuff it
	 * into the KolabMailAccess object. ( #3 )
	 */

	icalcomp = icalparser_parse_string (calobj);
	if (icalcomp == NULL) {
		g_warning (" + *calobj could not be parsed into an icalcomponent: %s",
		           calobj);
		return GNOME_Evolution_Calendar_InvalidObject;
	}
	/* now this better should be an vcalendar entry */
	if (icalcomponent_isa (icalcomp) != ICAL_VCALENDAR_COMPONENT) {
		g_warning (" + **calobj does not represent an vcalendar entry: %s",
		           calobj);
		icalcomponent_free (icalcomp);
		return GNOME_Evolution_Calendar_InvalidObject;
	}

	status = kolab_util_calendar_extract (icalcomp,
	                                      priv->source_type,
	                                      &ecalcomp,
	                                      &tzcomp);
	if (status != GNOME_Evolution_Calendar_Success) {
		icalcomponent_free (icalcomp);
		return status;
	}
	icalcomponent_free (icalcomp);

	kolab_util_calendar_store (ecalcomp,
	                           tzcomp,
	                           priv->default_zone,
	                           priv->cal_koma,
	                           priv->cal_uri,
	                           &error);
	if (error != NULL) {
		status = kolab_util_calendar_map_error (error);
		g_warning ("Error during KolabMailHandle creation: %s",
		           error->message);
		g_error_free (error);
		g_object_unref (ecalcomp);
		if (tzcomp != NULL)
			g_object_unref (tzcomp);
		return status;
	}

	if (tzcomp != NULL) {
		tzstr = e_cal_component_get_as_string (tzcomp);
		status = e_cal_backend_kolab_add_timezone (backend,
		                                           cal,
		                                           tzstr);
	}
	if (status != GNOME_Evolution_Calendar_Success)
		return status;
	e_cal_backend_notify_object_created (E_CAL_BACKEND (kolab),
	                                     e_cal_component_get_as_string (ecalcomp));
	g_object_unref (ecalcomp);
	if (tzcomp != NULL)
		g_object_unref (tzcomp);

	return status;
#endif
}

static void
e_cal_backend_kolab_send_objects (ECalBackendSync *backend,
                                  EDataCal *cal,
                                  GCancellable *cancellable,
                                  const gchar *calobj,
                                  GSList **users,
                                  gchar **modified_calobj,
                                  GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	(void)calobj; /* FIXME */
	(void)users; /* FIXME */ /* may be NULL? */
	(void)modified_calobj; /* FIXME */ /* must be NULL? */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

#if 0 /* FIXME old */
	(void) backend;
	(void) cal;

	*users = NULL;
	*modified_calobj = g_strdup (calobj);

	return GNOME_Evolution_Calendar_Success;
#endif
}

static void
e_cal_backend_kolab_get_attachment_uris (ECalBackendSync *backend,
                                         EDataCal *cal,
                                         GCancellable *cancellable,
                                         const gchar *uid,
                                         const gchar *rid,
                                         GSList **attachments,
                                         GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	g_assert (uid != NULL);
	(void)rid; /* FIXME */ /* may be NULL? */
	(void)attachments; /* FIXME */ /* must be NULL? */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);
}

static void
e_cal_backend_kolab_discard_alarm (ECalBackendSync *backend,
                                   EDataCal *cal,
                                   GCancellable *cancellable,
                                   const gchar *uid,
                                   const gchar *rid,
                                   const gchar *auid,
                                   GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	g_assert (uid != NULL);
	(void)rid; /* FIXME */ /* may be NULL? */
	(void)auid; /* FIXME */ /* may be NULL? */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);
}

static void
e_cal_backend_kolab_get_timezone (ECalBackendSync *backend,
                                  EDataCal *cal,
                                  GCancellable *cancellable,
                                  const gchar *tzid,
                                  gchar **tzobject,
                                  GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	g_assert (tzid != NULL);
	(void)tzobject; /* FIXME */ /* must be NULL? */
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);
}

/**
 * e_cal_backend_kolab_add_timezone:
 * @backend: An ECalBackendSync object.
 * @cal: An EDataCal object.
 * @cancellable: A cancellation stack.
 * @tzobject: VTIMEZONE object to be added.
 * @err: A GError placeholder.
 *
 * Add a timezone object to the given backend.
 */
static void
e_cal_backend_kolab_add_timezone (ECalBackendSync *backend,
                                  EDataCal *cal,
                                  GCancellable *cancellable,
                                  const gchar *tzobject,
                                  GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL (cal));
	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	g_assert (tzobject != NULL);
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

#if 0 /* FIXME old */
	(void) cal;

	icalcomp = icalparser_parse_string (tzobj);
	if (icalcomp == NULL)
		return GNOME_Evolution_Calendar_InvalidObject;

	if (icalcomponent_isa (icalcomp) == ICAL_VTIMEZONE_COMPONENT) {

		icaltz = icaltimezone_new ();
		icaltimezone_set_component (icaltz,
		                            icalcomp);

		if (e_cal_backend_cache_put_timezone (priv->cal_cache, icaltz) == FALSE) {
			g_warning ("%s()[%u]: Putting timezone in cache did not succeed.",
			           __func__, __LINE__);
			icaltimezone_free (icaltz, 1);
			return GNOME_Evolution_Calendar_OtherError;
		}
		icaltimezone_free (icaltz, 1);
	}
	return GNOME_Evolution_Calendar_Success;
#endif
}

static void
e_cal_backend_kolab_authenticate_user (ECalBackendSync *backend,
                                       GCancellable *cancellable,
                                       ECredentials *credentials,
                                       GError **err)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	(void)cancellable; /* FIXME */ /* cancellable may be NULL */
	g_assert (credentials != NULL);
	g_return_if_fail (err == NULL || *err == NULL);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);
}

/**
 * e_cal_backend_kolab_start_view:
 * @backend: A calendar backend.
 * @view: The view to be started.
 *
 * Starts a new live query on the given backend.
 */
static void
e_cal_backend_kolab_start_view (ECalBackend *backend,
                                EDataCalView *view)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL_VIEW (view));

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);

#if 0 /* FIXME old */
	query = e_data_cal_view_get_text (cal_view);
	ksexp = e_cal_backend_sexp_new (query);

	sourcename = kolab_util_backend_get_relative_path_from_uri (priv->cal_uri);

	kolab_util_calendar_cache_update_on_query (priv->cal_cache,
	                                           priv->cal_koma,
	                                           query,
	                                           priv->cal_uri);
	uid_list = kolab_mail_access_query_uids (priv->cal_koma,
	                                         sourcename,
	                                         query,
	                                         &error);
	if (error != NULL) {
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__, error->message);
		g_error_free (error);
		error = NULL;
	}

	for (it = g_list_first (uid_list); it != NULL; it = g_list_next (it)) {
		gchar *uid = it->data;
		ECalComponent *ecalcomp = NULL;
		ecalcomp = kolab_util_calendar_cache_get_object (priv->cal_cache,
		                                                 priv->cal_koma,
		                                                 priv->cal_uri,
		                                                 uid,
		                                                 TRUE,
		                                                 &error);
		if (error != NULL) {
			g_warning ("%s()[%u]: %s",
			           __func__, __LINE__, error->message);
			g_error_free (error);
			error = NULL;
		}
		if (ecalcomp != NULL) {
			iCal_objects = g_list_append (iCal_objects,
			                              e_cal_component_get_as_string (ecalcomp));
			g_object_unref (ecalcomp);
		}
	}

	e_data_cal_view_notify_objects_added (cal_view,
	                                      iCal_objects);

	g_list_foreach (iCal_objects, (GFunc) g_free, NULL);
	g_list_free (iCal_objects);
	g_object_unref (ksexp);
	g_free (sourcename);

	kolab_util_glib_glist_free (uid_list);

	e_data_cal_view_notify_done (cal_view, status);
#endif
}

static void
e_cal_backend_kolab_stop_view (ECalBackend *backend,
                               EDataCalView *view)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_assert (E_IS_DATA_CAL_VIEW (view));

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);
}

static void
e_cal_backend_kolab_last_client_gone (ECalBackend *backend)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	(void)self;
	(void)priv;
	g_error ("%s: FIXME implement me", __func__);
}

/*----------------------------------------------------------------------------*/
/* object/class init */

static void
e_cal_backend_kolab_init (ECalBackendKolab *backend)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_debug ("%s()[%u] called.", __func__, __LINE__);

	priv->koma_table = NULL;
	priv->cal_mode = 0;
	priv->cal_koma = NULL;
	priv->cal_cache = NULL;
	priv->user_email = NULL;
	priv->default_zone = NULL;
	priv->source_type = 0;
	priv->already_opened = FALSE;
	priv->cal_uri = NULL;
} /* e_cal_backend_kolab_init () */

static void
e_cal_backend_kolab_dispose (GObject *object)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (object);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_debug ("%s()[%u] called.", __func__, __LINE__);

	if (priv->already_opened != FALSE) {
		priv->already_opened = FALSE;
		g_object_unref (priv->cal_koma);
		priv->cal_koma = NULL;
	}
	if (priv->default_zone != NULL) {
		g_object_unref (priv->default_zone);
		priv->default_zone = NULL;
	}
	if (priv->cal_cache != NULL) {
		g_object_unref (priv->cal_cache);
		priv->cal_cache = NULL;
	}

	G_OBJECT_CLASS (e_cal_backend_kolab_parent_class)->dispose (object);
} /* e_cal_backend_kolab_dispose () */

static void
e_cal_backend_kolab_finalize (GObject *object)
{
	ECalBackendKolab *self = E_CAL_BACKEND_KOLAB (object);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (self);

	g_debug ("%s()[%u] called.", __func__, __LINE__);

	if (priv->user_email != NULL)
		g_free (priv->user_email);
	if (priv->cal_uri != NULL)
		g_free (priv->cal_uri);

	G_OBJECT_CLASS (e_cal_backend_kolab_parent_class)->finalize (object);
} /* e_book_backend_kolab_finalize () */

/**
 * e_cal_backend_kolab_class_init:
 * @klass: An ECalBackendKolabClass type.
 *
 * Class initialisation function for the Kolab backend.
 */
static void
e_cal_backend_kolab_class_init (ECalBackendKolabClass *klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);
	ECalBackendClass *backend_class = E_CAL_BACKEND_CLASS (klass);
	ECalBackendSyncClass* sync_class = E_CAL_BACKEND_SYNC_CLASS (klass);

	g_type_class_add_private (klass, sizeof (ECalBackendKolabPrivate));

	/* Sync backend class functions */
	sync_class->open_sync = e_cal_backend_kolab_open;
	sync_class->remove_sync = e_cal_backend_kolab_remove;
	sync_class->refresh_sync = e_cal_backend_kolab_refresh;
	sync_class->get_backend_property_sync = e_cal_backend_kolab_get_backend_property;
	sync_class->set_backend_property_sync = e_cal_backend_kolab_set_backend_property;
	sync_class->get_object_sync = e_cal_backend_kolab_get_object;
	sync_class->get_object_list_sync = e_cal_backend_kolab_get_object_list;
	sync_class->get_free_busy_sync = e_cal_backend_kolab_get_free_busy;
	sync_class->create_object_sync = e_cal_backend_kolab_create_object;
	sync_class->modify_object_sync = e_cal_backend_kolab_modify_object;
	sync_class->remove_object_sync = e_cal_backend_kolab_remove_object;
	sync_class->receive_objects_sync = e_cal_backend_kolab_receive_objects;
	sync_class->send_objects_sync = e_cal_backend_kolab_send_objects;
	sync_class->get_attachment_uris_sync = e_cal_backend_kolab_get_attachment_uris;
	sync_class->discard_alarm_sync = e_cal_backend_kolab_discard_alarm;
	sync_class->get_timezone_sync = e_cal_backend_kolab_get_timezone;
	sync_class->add_timezone_sync = e_cal_backend_kolab_add_timezone;
	sync_class->authenticate_user_sync = e_cal_backend_kolab_authenticate_user;

	/* Backend parent class methods methods not covered in the sync backend part */
	backend_class->start_view = e_cal_backend_kolab_start_view;
	backend_class->stop_view = e_cal_backend_kolab_stop_view;
	backend_class->last_client_gone = e_cal_backend_kolab_last_client_gone;

	object_class->dispose = e_cal_backend_kolab_dispose;
	object_class->finalize = e_cal_backend_kolab_finalize;

} /* e_cal_backend_kolab_class_init () */


/*----------------------------------------------------------------------------*/

#if 0 /* FIXME old */

/**
 * e_cal_backend_kolab_get_default_object:
 * @backend: An ECalBackendSync object.
 * @cal: An EDataCal object.
 * @object: Placeholder for returned object.
 *
 * Returns the default object via object placeholder.
 *
 * Returns: Status code.
 */
static ECalBackendSyncStatus
e_cal_backend_kolab_get_default_object (ECalBackendSync *backend,
                                        EDataCal *cal,
                                        gchar **object)
{
	/* FIXME: Remove experimental code related to summary*/
	ECalComponent *comp;
	ECalComponentText *summary = NULL;
	ECalComponentText *desc = NULL;
	GSList *desclist = NULL;
	gchar *tmpsum = g_strdup ("");
	gchar *tmploc = g_strdup ("");
	gchar *tmpdesc1 = g_strdup ("");
	(void) cal;

	summary = g_new0 (ECalComponentText, 1);
	summary->value=tmpsum;
	desc = g_new0 (ECalComponentText, 1);
	desc->value=tmpdesc1;
	desclist = g_slist_append(desclist, desc);

	comp = e_cal_component_new ();

	switch (e_cal_backend_get_kind (E_CAL_BACKEND (backend))) {
	case ICAL_VEVENT_COMPONENT:
		e_cal_component_set_new_vtype (comp, E_CAL_COMPONENT_EVENT);
		e_cal_component_set_summary (comp, summary);
		e_cal_component_set_description_list (comp, desclist);
		e_cal_component_set_location (comp, tmploc);
		break;
	default:
		g_object_unref (comp);
		return GNOME_Evolution_Calendar_ObjectNotFound;
	}

	*object = e_cal_component_get_as_string (comp);

	if (comp)
		g_object_unref (comp);

	g_free (summary);
	g_free (tmploc);
	g_free (tmpdesc1);
	g_slist_foreach (desclist, (GFunc) g_free, NULL);
	g_slist_free (desclist);

	return GNOME_Evolution_Calendar_Success;
} /* e_cal_backend_kolab_get_default_object () */

/**
 * e_cal_backend_kolab_set_default_zone:
 * @backend: An ECalBackendSync object.
 * @cal: An EDataCal object.
 * @tz: Timezone object as string.
 *
 * Sets the default timezone for the calendar, which is used to resolve DATE and
 * floating DATE-TIME values.
 *
 * Returns: Status code.
 */
static ECalBackendSyncStatus
e_cal_backend_kolab_set_default_zone (ECalBackendSync *backend,
                                      EDataCal *cal,
                                      const gchar *tzobj)
{
	ECalBackendKolab *kolab = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (kolab);
	ECalComponent *ecaltz = NULL;
	(void) cal;

	ecaltz = e_cal_component_new_from_string (tzobj);
	if (ecaltz == NULL)
		return GNOME_Evolution_Calendar_InvalidObject;

	if (priv->default_zone != NULL)
		g_object_unref (priv->default_zone);
	priv->default_zone = ecaltz;
	return GNOME_Evolution_Calendar_Success;
} /* e_cal_backend_kolab_set_default_zone () */

/**
 * e_cal_backend_internal_get_default_timezone:
 * @backend: A calendar backend.
 *
 * Requests the default timezone from the backend.
 *
 * Returns: The icaltimezone representation of the default timezone.
 */
static icaltimezone*
e_cal_backend_kolab_internal_get_default_timezone (ECalBackend *backend)
{
	ECalBackendKolab *kolab = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (kolab);
	icaltimezone *icaltz = NULL;
	icalcomponent *icalcomp = NULL;

	icaltz = icaltimezone_new ();
	icalcomp = e_cal_component_get_icalcomponent (priv->default_zone);
	if (icalcomp == NULL)
		return NULL;
	icaltimezone_set_component (icaltz, icalcomp);

	return icaltz;
} /* e_cal_backend_kolab_internal_get_default_timezone () */

/**
 * e_cal_backend_kolab_get_mode:
 * @backend: A Kolab calendar backend.
 *
 * Queries the current calendar backend mode (local/remote).
 *
 * Returns: The current mode, the calendar is in.
 */
static CalMode
e_cal_backend_kolab_get_mode (ECalBackend *backend)
{
	ECalBackendKolab *kolab = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (kolab);
	KolabMailAccessOpmodeID koma_mode = KOLAB_MAIL_ACCESS_OPMODE_INVAL;
	GError *error = NULL;

	if (priv->cal_koma == NULL) {
		g_warning ("%s()[%u]: KolabMailAccess object not existent.",
		           __func__, __LINE__);
		return CAL_MODE_INVALID;
	}
	koma_mode = kolab_mail_access_get_opmode (priv->cal_koma, &error);
	if (error != NULL) {
		g_warning ("%s()[%u]: %s",
		           __func__, __LINE__, error->message);
		g_error_free (error);
		return CAL_MODE_INVALID;
	}

	switch (koma_mode) {
	case KOLAB_MAIL_ACCESS_OPMODE_OFFLINE:
		return CAL_MODE_LOCAL;
		break;
	case KOLAB_MAIL_ACCESS_OPMODE_ONLINE:
		return CAL_MODE_REMOTE;
		break;
	case KOLAB_MAIL_ACCESS_OPMODE_SHUTDOWN:
	case KOLAB_MAIL_ACCESS_OPMODE_NEW:
	case KOLAB_MAIL_ACCESS_OPMODE_INVAL:
	case KOLAB_MAIL_ACCESS_OPMODE_CONFIGURED:
	default:
		g_warning ("%s()[%u]: KolabMailAccess object's state unknown: %u.",
		           __func__, __LINE__, koma_mode);
		return CAL_MODE_INVALID;
	}
} /* e_cal_backend_kolab_get_mode () */

/**
 * e_cal_backend_kolab_set_mode:
 * @backend: A Kolab calendar backend.
 * @mode: Mode to change to.
 *
 * Sets the mode of the calendar backend to the given mode.
 */
static void
e_cal_backend_kolab_set_mode (ECalBackend *backend,
                              CalMode mode)
{
	ECalBackendKolab *kolab = E_CAL_BACKEND_KOLAB (backend);
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (kolab);
	GError *error = NULL;
	KolabMailAccessOpmodeID koma_mode = KOLAB_MAIL_ACCESS_OPMODE_INVAL;
	KolabMailAccessOpmodeID tmp_mode = KOLAB_MAIL_ACCESS_OPMODE_INVAL;
	gchar *sourcename = NULL;
	KolabSettingsHandler *ksettings = NULL;
	gboolean ok;

	switch (mode) {
	case CAL_MODE_LOCAL:
		priv->cal_mode = CAL_MODE_LOCAL;
		koma_mode = KOLAB_MAIL_ACCESS_OPMODE_OFFLINE;
		break;
	case CAL_MODE_REMOTE:
		priv->cal_mode = CAL_MODE_REMOTE;
		koma_mode = KOLAB_MAIL_ACCESS_OPMODE_ONLINE;
		break;
	case CAL_MODE_ANY:
	case CAL_MODE_INVALID:
	default:
		g_assert_not_reached();
	}
	if (priv->cal_koma == NULL) {
		g_warning ("%s()[%u]: KolabMailAccess object not existent.",
		           __func__, __LINE__);
		return;
	}
	tmp_mode = kolab_mail_access_get_opmode (priv->cal_koma,
	                                         &error);
	if (error != NULL)
		{
			g_warning ("%s()[%u]: error getting mode: %s",
			           __func__, __LINE__, error->message);
			g_error_free (error);
			return;
		}

	if (tmp_mode < KOLAB_MAIL_ACCESS_OPMODE_OFFLINE) {
		return;
	}
	ok = kolab_mail_access_set_opmode (priv->cal_koma,
	                                   koma_mode,
	                                   &error);
	if (error != NULL)
		{
			g_warning ("%s()[%u]: error setting mode: %s",
			           __func__, __LINE__, error->message);
			g_error_free (error);
			error = NULL;
			return;
		}

	if (ok && (koma_mode < KOLAB_MAIL_ACCESS_OPMODE_ONLINE))
		return;

	sourcename = kolab_util_backend_get_relative_path_from_uri (priv->cal_uri);
	ok = kolab_mail_access_source_fbtrigger_needed (priv->cal_koma,
	                                                sourcename,
	                                                &error);
	if (error != NULL)
		{
			g_warning ("%s()[%u]: %s",
			           __func__, __LINE__, error->message);
			g_error_free (error);
			error = NULL;
			goto cleanup;
		}
	if (ok) {
		ksettings = kolab_mail_access_get_settings_handler (priv->cal_koma);
		kolab_util_calendar_toggle_pfb_trigger (ksettings,
		                                        sourcename);
		g_object_unref (ksettings);
	}
 cleanup:
	g_free (sourcename);
} /* e_cal_backend_kolab_set_mode () */

/**
 * e_cal_backend_kolab_set_koma_table:
 * @kolab: An ECalBackendKolab object.
 * @koma_objects: A GHashTable to contain the required KoMA instances.
 *
 * This method has to be called before any other method which accesses
 * Kolab infrastructure. In this case it should be called from
 * e-cal-backend-kolab-factory during creation of a new ECalBackendKolab
 * instance.
 */
void
e_cal_backend_kolab_set_koma_table (ECalBackendKolab *kolab,
                                    GHashTable *koma_objects)
{
	ECalBackendKolabPrivate *priv = E_CAL_BACKEND_KOLAB_PRIVATE (kolab);

	g_hash_table_ref (koma_objects);
	priv->koma_table = koma_objects;
}

#endif
